#lang errortrace racket


(define (copy l)
  (match l ; (list 1 2 3)
    [(list) (list)]
    [(list h l ...)
      ; h = 1
      ; l = (list 2 3)
      (define result (copy l))
      ; result = (list 2 3)
      (cons h result)
      ; output = (list 1 2 3)
      ]
  ))

(define (copy-foldr l)
  (foldr
    ; step
;;    cons
    (lambda (h result)
      (displayln
        (format "step: h=~a result=~a"
          h result))
      (cons h result)
    )
    ; base-case
    (list)
    ; list we are iterating over
    l
    )
)


(define (copy-foldl l)
  (foldl
    ; step
;;    cons
    (lambda (h result)
      (displayln
        (format "step: h=~a result=~a"
          h result))
      (cons h result)
    )
    ; base-case
    (list)
    ; list we are iterating over
    l
    )
)

; (copy-foldl (list 1 2 3))
; (list 3 2 1)

(define (cons-right x l)
  ; x = 4
  ; l = (list 1 2 3)
  (match l
    [(list) (list x)]
    [(list h l ...)
      ; h = 1
      ; l = (list 2 3)
      (define result (cons-right x l))
      ; result = (list 2 3 4)
      (cons h result)
      ; output = (list 1 2 3 4)
    ]
    ))

(cons-right 4 (list 1 2 3))
; (list 1 2 3 4)

(define (reverse-slow l)
  (match l ; (list 1 2 3)
    [(list) (list)]
    [(list h l ...)
      ; h = 1
      ; l = (list 2 3)
      (define result (reverse-slow l))
      ; result = (list 3 2)
      (cons-right h result)
    ]
  )
  )

;(reverse-slow (list 1 2 3 4))
(define (reverse l)
  (define (reverse-iter rev-list l)
    ; rev-list = (list 2 1)
    ; l = (list 3 4)
    (match l
      [(list) rev-list]
      [(list h l ...)
        (displayln (format
          "step: h=~a l=~a rev-list=~a"
          h l rev-list))
        ; h = 3
        ; l = (list 4)
        ; ----------
        ; h = 4
        ; l = (list)
        (define new-rev-list (cons h rev-list))
        ; new-rev-list = (list 3 2 1)
        ; --------
        ; new-rev-list = (list 4 3 2 1)
        (reverse-iter new-rev-list l)
      ]
    )
  )
  (reverse-iter (list) l))
(reverse (list 1 2 3 4))
(define (join-suffix sep l)
  (match l
    [(list) ""]
    [(list x) x]
    [(list h l ...)
      (string-append h sep (join-suffix sep l))
    ]
  )
)
#|
(join-suffix ", " (list "a" "b" "c"))
(join-suffix ", " (list "a" "b"))
(join-suffix ", " (list "a"))
(join-suffix ", " (list))
|#
(define (join-prefix sep l)
  (match l
    [(list) ""]
    [(list h l ...)
      (define (join-iter l)
        (match l
          [(list) ""]
          [(list h l ...)
            ; h = "b"
            ; l = (list "c")
            (string-append sep h
              (join-iter l))
            ; output = ", b, c"
          ]
        )
      )
      (string-append h (join-iter l))
    ]
  )
)

(join-prefix ", " (list "a" "b" "c"))
(join-prefix ", " (list "a" "b"))
(join-prefix ", " (list "a"))
(join-prefix ", " (list))

(define (join-foldr sep l)
  (match l
    [(list) ""]
    [(list h l ...)
      (define (join-iter l)
        (foldr
          (lambda (h result)
            (string-append sep h
              result)
          )
          ""
          l)
      )
      (string-append h (join-iter l))
    ]
  )
)

(join-foldr ", " (list "a" "b" "c"))
(join-foldr ", " (list "a" "b"))
(join-foldr ", " (list "a"))
(join-foldr ", " (list))

(define (join-fast sep l)
  (match l
    [(list) ""]
    [(list h l ...)
      (define (join-iter accum l)
        (match l
          [(list) accum]
          [(list h l ...)
            ; h = "c"
            ; l = (list)
            ; accum = ", b"
            (define new-accum (string-append accum sep h))
            (join-iter new-accum l)
          ]
        )
      )
      (string-append h (join-iter "" l))
    ]
  )
)


(join-fast ", " (list "a" "b" "c"))
(join-fast ", " (list "a" "b"))
(join-fast ", " (list "a"))
(join-fast ", " (list))

(define (join-foldl sep l)
  (match l
    [(list) ""]
    [(list h l ...)
;;       (define (join-iter accum l)
;;         (foldr ""
;;         (match l
;;           [(list) accum]
;;           [(list h l ...)
;;             ; h = "c"
;;             ; l = (list)
;;             ; accum = ", b"
;;             (define new-accum (string-append accum sep h))
;;             (join-iter new-accum l)
;;           ]
;;         )
;;       )
      (string-append h (foldr
        (lambda (h new-accum)
          (string-append new-accum sep h))
        "" l))
    ]
  )
)


(join-foldl ", " (list "a" "b" "c"))
(join-foldl ", " (list "a" "b"))
(join-foldl ", " (list "a"))
(join-foldl ", " (list))



